# This file is part of the lib3to6 project
# https://gitlab.com/mbarkhau/lib3to6
#
# Copyright (c) 2019 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT

import ast
import typing as typ

from . import common

# NOTE (mb 2018-06-24): Version info pulled from:
# https://docs.python.org/3/library/__future__.html


class VersionInfo:

    apply_since: typ.List[int]
    apply_until: typ.List[int]
    works_since: typ.List[int]
    works_until: typ.Optional[typ.List[int]]

    def __init__(
        self, apply_since: str, apply_until: str, works_since: str = None, works_until: str = None
    ) -> None:

        self.apply_since = [int(part) for part in apply_since.split(".")]
        self.apply_until = [int(part) for part in apply_until.split(".")]
        if works_since is None:
            # Implicitly, if it's applied since a version, it
            # also works since then.
            self.works_since = self.apply_since
        else:
            self.works_since = [int(part) for part in works_since.split(".")]
        self.works_until = [int(part) for part in works_until.split(".")] if works_until else None


class FixerBase:

    version_info       : VersionInfo
    required_imports   : typ.Set[common.ImportDecl]
    module_declarations: typ.Set[str]

    def __init__(self) -> None:
        self.required_imports    = set()
        self.module_declarations = set()

    def __call__(self, cfg: common.BuildConfig, tree: ast.Module) -> ast.Module:
        raise NotImplementedError()

    @classmethod
    def is_required_for(cls, version: str) -> bool:
        version_num = [int(part) for part in version.split(".")]
        nfo         = cls.version_info
        return nfo.apply_since <= version_num <= nfo.apply_until

    @classmethod
    def is_compatible_with(cls, version: str) -> bool:
        version_num = [int(part) for part in version.split(".")]
        nfo         = cls.version_info
        return nfo.works_since <= version_num and (
            nfo.works_until is None or version_num <= nfo.works_until
        )

    @classmethod
    def is_applicable_to(cls, src_version: str, tgt_version: str) -> bool:
        return cls.is_compatible_with(src_version) and cls.is_required_for(tgt_version)


class TransformerFixerBase(FixerBase, ast.NodeTransformer):
    def __call__(self, cfg: common.BuildConfig, tree: ast.Module) -> ast.Module:
        try:
            return self.visit(tree)
        except common.FixerError as ex:
            if ex.module is None:
                ex.module = tree
            raise



#!/usr/bin/env python
# This file is part of the lib3to6 project
# https://gitlab.com/mbarkhau/lib3to6
#
# Copyright (c) 2019 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT

import io
import os
import re
import sys
import typing as typ
import difflib

import click

from . import common
from . import packaging
from . import transpile

# To enable pretty tracebacks:
#   echo "export ENABLE_BACKTRACE=1;" >> ~/.bashrc
if os.environ.get('ENABLE_BACKTRACE') == '1':
    import backtrace

    backtrace.hook(align=True, strip_path=True, enable_on_envvar_only=True)


click.disable_unicode_literals_warning = True


def _print_diff(source_text: str, fixed_source_text: str) -> None:
    differ = difflib.Differ()

    source_lines       = source_text.splitlines()
    fixed_source_lines = fixed_source_text.splitlines()
    diff_lines         = differ.compare(source_lines, fixed_source_lines)
    if not sys.stdout.isatty():
        click.echo("\n".join(diff_lines))
        return

    for line in diff_lines:
        if line.startswith("+ "):
            click.echo("\u001b[32m" + line + "\u001b[0m")
        elif line.startswith("- "):
            click.echo("\u001b[31m" + line + "\u001b[0m")
        elif line.startswith("? "):
            click.echo("\u001b[36m" + line + "\u001b[0m")
        else:
            click.echo(line)
    print()


@click.command()
@click.option(
    "--target-version", default="2.7", metavar="<version>", help="Target version of python."
)
@click.option(
    "--diff", default=False, is_flag=True, help="Output diff instead of transpiled source."
)
@click.option("--in-place", default=False, is_flag=True, help="Write result back to input file.")
# @click.option(
#     "--config",
#     default="lib3to6.toml",
#     required=False,
#     metavar="<path>",
#     help="Path to config file.",
# )
@click.argument("source_files", metavar="<source_file>", nargs=-1, type=click.File(mode="r"))
def main(
    target_version: str,
    diff          : bool,
    in_place      : bool,
    # config        : str,
    source_files: typ.Sequence[io.TextIOWrapper],
) -> None:
    if not any(source_files):
        print("No files.")
        sys.exit(1)

    if target_version and not re.match(r"[0-9]+\.[0-9]+", target_version):
        print(f"Invalid argument --target-version={target_version}")
        sys.exit(1)

    cfg = packaging.eval_build_config(target_version=target_version)
    for src_file in source_files:
        source_text = src_file.read()
        try:
            fixed_source_text = transpile.transpile_module(cfg, source_text)
        except common.CheckError as err:
            err.args = (err.args[0] + f" of file {src_file.name} ",) + err.args[1:]
            raise

        if diff:
            _print_diff(source_text, fixed_source_text)
        elif in_place:
            with io.open(src_file.name, mode="w") as fh:
                fh.write(fixed_source_text)
        else:
            print(fixed_source_text)


if __name__ == '__main__':
    main()  # type: ignore
